from .common import (  # exported
    Message,
    StdioMessage,
    AuthValidJWTMessage,
    AuthInvalidJWTMessage,
    AuthRequireInfoMessage,
)


# class ProgramEndMessage(Message):
#     type = "program:end"

#     def __init__(self, exit_code: int, exception: str, frames: list):
#         super().__init__(
#             {
#                 "exitCode": exit_code,
#                 "exception": exception,
#             }
#         )


class SessionIdMessage(Message):
    type = "session-id"

    def __init__(self, session_id: str):
        super().__init__({"sessionId": session_id})


class GenericMessage(Message):
    def __init__(self, data: dict):
        self.type = data["type"]
        super().__init__(data)



from .example_instances import example_instances
from abstra.widgets.metadata_parsed import metadata_objects
from abstra.widgets.validation import validate_widget_props

generated_widgets = [example_instance.json() for example_instance in example_instances]
del metadata_objects["list-input"]


class TestFunction:
    def test_widget_examples_contains_all_widgets(self):
        for widget_type in metadata_objects:
            if widget_type != "list-input":
                assert any(
                    widget["type"] == widget_type
                    for widget in generated_widgets
                    if widget["type"] != "list-input"
                )

    def test_widgets_exist_in_metadata(self):
        for widget_example in generated_widgets:
            assert widget_example["type"] in metadata_objects

    def test_widgets_props_are_right(self):
        for widget_example in generated_widgets:
            validate_widget_props(widget_example)



import webbrowser, sys, os, json
from glob import glob

from ..resources.resources import Resource
from .. import messages as messages
from .. import utils as utils
from .. import apis as apis


NAME_PARAMETERS = ["name", "title"]
PATH_PARAMETERS = ["path"]
CODE_PARAMETERS = ["code", "c"]
BACKGROUND_PARAMETERS = ["background"]
OTHER_PARAMETERS = [
    "main_color",
    "font_family",
    "logo_url",
    "log_messages",
    "brand_name",
]

LAYOUT_PARAMETERS = ["layout"]
NON_FLAG_PARAMETERS = (
    NAME_PARAMETERS
    + PATH_PARAMETERS
    + CODE_PARAMETERS
    + BACKGROUND_PARAMETERS
    + OTHER_PARAMETERS
    + LAYOUT_PARAMETERS
)
FLAG_PARAMETERS = ["enabled"]
DASH_PARAMETERS = FLAG_PARAMETERS + NON_FLAG_PARAMETERS


def check_valid_parameters(parameters: dict) -> None:
    for param in parameters.keys():
        if param not in DASH_PARAMETERS:
            messages.invalid_parameter(param)
            sys.exit(1)
    for param, value in parameters.items():
        if param in NON_FLAG_PARAMETERS and value in [True, False]:
            messages.invalid_non_flag_parameter_value(param)
            sys.exit(1)


def evaluate_parameter_title(parameters: dict, use_default=True) -> dict:
    title = parameters.get("name") or parameters.get("n") or parameters.get("title")
    if not title and not use_default:
        return {}
    return {"title": title or "New Dash"}


def evaluate_optional_parameter(parameter_name: str, parameters: dict) -> dict:
    parameter_value = parameters.get(parameter_name)
    if parameter_value is None:
        return {}
    return {parameter_name: parameter_value}


def evaluate_parameters_code(parameters: dict) -> dict:
    code = parameters.get("code") or parameters.get("c")

    if not code:
        print("Code is required")
        sys.exit(1)

    return {"code_file_path": code}


def evaluate_parameter_layout(parameters: dict) -> dict:
    layout = parameters.get("layout")
    if not layout:
        print("Layout is required")
        sys.exit(1)

    return {"layout": layout}


def evaluate_flag_parameters(parameters: dict) -> dict:
    evaluated_params = {}
    for param, value in parameters.items():
        if param in FLAG_PARAMETERS:
            if value == "true" or value == True:
                evaluated_params[param] = True
                continue
            if value == "false" or value == False:
                evaluated_params[param] = False
                continue

            messages.invalid_flag_parameter_value(param)
            sys.exit(1)
    return evaluated_params


def evaluate_other_parameters(parameters: dict) -> dict:
    other_parameters = {}
    for param, value in parameters.items():
        if param in OTHER_PARAMETERS:
            other_parameters[param] = value

    return other_parameters


def evaluate_background_parameter_value(parameters: dict) -> dict:
    background = parameters.get("background", None)
    if not background:
        return {}

    if utils.check_color(background):
        return {"theme": background}

    if utils.check_is_url(background):
        return {"theme": background}

    if not utils.path_exists(background):
        messages.file_path_does_not_exists_message(background)
        sys.exit(1)

    if utils.check_is_image_path(background):
        filename = utils.slugify_filename(background)
        try:
            with open(background, "rb") as f:
                file = f.read()
                url = apis.asset_upload(filename, file)
                return {"theme": url}
        except Exception as e:
            messages.error_upload_background_message(background)
            sys.exit(1)

    messages.invalid_background_parameter_value()
    sys.exit(1)


def upsert_dash_data_from_kwargs(kwargs: dict) -> dict:
    return {
        **evaluate_parameters_code(kwargs),
        **evaluate_flag_parameters(kwargs),
        **evaluate_other_parameters(kwargs),
        **evaluate_background_parameter_value(kwargs),
        **evaluate_parameter_layout(kwargs),
        **evaluate_optional_parameter("path", kwargs),
        **evaluate_optional_parameter("main_color", kwargs),
        **evaluate_optional_parameter("font_family", kwargs),
        **evaluate_optional_parameter("logo_url", kwargs),
        **evaluate_optional_parameter("brand_name", kwargs),
    }


class Dashes(Resource):
    @staticmethod
    def list():
        dashes = apis.list_workspace_dashes()
        messages.print_dashes(dashes)

    @staticmethod
    def add(*args, **kwargs):
        upsert = kwargs.pop("upsert", False)
        path = kwargs.get("path")
        if upsert and path is None:
            messages.upsert_without_identifier("path")
            sys.exit(1)

        check_valid_parameters(kwargs)

        dash_data = {
            **evaluate_parameter_title(kwargs),
            **upsert_dash_data_from_kwargs(kwargs),
        }

        if dash_data:
            try:
                if upsert:
                    apis.upsert_workspace_dash(dash_data)
                    messages.upserted_message("Dash", path or "main")
                else:
                    path = apis.add_workspace_dash(dash_data)["path"]
                    messages.created_message("Dash", path or "main")
            except Exception as e:
                print(e)
                messages.create_failed("Dash")
                sys.exit(1)

    @staticmethod
    def update(*args, **kwargs):
        if not len(args):
            messages.required_argument("path")
            sys.exit(1)
        path = args[0]

        if not len(kwargs):
            messages.missing_parameters_to_update("dash", path)
            sys.exit(1)

        check_valid_parameters(kwargs)

        dash_data = {
            **evaluate_parameter_title(kwargs, use_default=False),
            **upsert_dash_data_from_kwargs(kwargs),
        }

        if dash_data:
            try:
                apis.update_workspace_dash(path, dash_data)
                messages.updated_message("Dash", path)
            except Exception as e:
                print(e)
                messages.update_failed("Dash", path)
                sys.exit(1)

    @staticmethod
    def remove(*args, **kwargs):
        if not len(args):
            messages.required_argument("path")
            sys.exit(1)

        path = args[0]
        apis.delete_workspace_dash(path)
        messages.deleted_message("Dash", path)

    @staticmethod
    def play(*args, **kwargs):
        if not len(args):
            messages.required_argument("path")
            sys.exit(1)

        path = args[0]
        subdomain_name = apis.get_subdomain()
        url = utils.get_prod_dash_url(subdomain_name, path)
        messages.dash_url(url)
        webbrowser.open(url)

    @staticmethod
    def map_deploy_data(abstra_json_path: str, abstra_json_data: dict):
        workspace_data = abstra_json_data.get("workspace")
        if not workspace_data:
            return []

        abstra_json_dir = os.path.dirname(abstra_json_path)
        base_path = utils.remove_suffix(workspace_data.get("base_path", ""), "/")

        dash_files = glob(
            os.path.join(abstra_json_dir, "**", "*.abstradash.json"),
            recursive=True,
        )
        dash_props = []
        for dash_file_path in dash_files:
            common_path = dash_file_path.replace(".abstradash.json", "")
            route = os.path.relpath(
                common_path, os.path.join(abstra_json_dir, ".")
            ).replace("\\", "/")

            route = utils.remove_prefix(route, f"{base_path}/")

            script_path = common_path + ".py"
            dash_json_data = json.load(open(dash_file_path, "r"))
            prop = {
                "title": dash_json_data.get("title") or route,
                "layout": dash_json_data["layout"],
                "background": workspace_data.get("theme"),
                "main_color": workspace_data.get("main_color"),
                "font_family": workspace_data.get("font_family"),
                "brand_name": workspace_data.get("brand_name"),
                "logo_url": workspace_data.get("logo_url"),
                "path": route if route != "main" else "",
                "code": script_path,
            }
            dash_props.append(prop)
        return dash_props

    @staticmethod
    def logs(*args, **kwargs):
        path = kwargs.pop("path", None)
        limit = kwargs.pop("limit", 20)
        offset = kwargs.pop("offset", 0)

        if limit == 0:
            limit = None
        if path is None:
            logs = apis.dashes.list_logs(limit=limit, offset=offset)
        else:
            logs = apis.dashes.list_logs_by_path(path, limit=limit, offset=offset)

        serialized_logs = json.dumps(logs, default=str, indent=4)
        messages.print_logs(serialized_logs)



import socket
from contextlib import closing


def get_free_port(default_port=3000):
    for port in range(default_port, default_port + 10):
        with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
            port_is_available = sock.connect_ex(("localhost", port)) != 0
            if port_is_available:
                return port



import tempfile
import unittest
from uuid import uuid4 as uuid
from pathlib import Path

from abstra_server.api import API

from .fixtures import init_dir, clear_dir


class TestWorkspace(unittest.TestCase):
    def setUp(self) -> None:
        self.root = init_dir()

    def tearDown(self) -> None:
        clear_dir(self.root)

    def test_api_update(self):
        api = API(root=self.root)

        api.update_workspace({"name": "test-workspace-updated"})

        workspace = api.get_workspace()

        self.assertEqual(workspace.name, "test-workspace-updated")



import re, typing, sqlite3


class TableNotFound(Exception):
    def __init__(self, table_name: str) -> None:
        self.table_name = table_name

    def __str__(self) -> str:
        return f"Table '{self.table_name}' does not exist"


class MissingColumns(Exception):
    def __init__(self, missing_columns: typing.List[str]) -> None:
        self.missing_columns = missing_columns

    def __str__(self) -> str:
        return f"Missing columns: {self.missing_columns}"


class InvalidTableJoin(Exception):
    def __init__(self, table) -> None:
        self.table = table

    def __str__(self) -> str:
        return f"Invalid table argument for join: {self.table}"


class SqliteColumn:
    table: "SqliteTable"
    id: int
    name: str
    type: str
    not_null: bool
    primary_key: bool
    default: typing.Optional[str] = None

    def __init__(
        self,
        table: "SqliteTable",
        id: int,
        name: str,
        type: str,
        not_null: bool,
        primary_key: bool,
        default: typing.Optional[str] = None,
    ) -> None:
        self.table = table
        self.id = id
        self.name = name
        self.type = type
        self.not_null = not_null
        self.primary_key = primary_key
        self.default = default

    @property
    def editor_dto(self):
        return {
            "id": self.id,
            "name": self.name,
            "type": self.type,
            "not_null": self.not_null,
            "default": self.default,
            "primary_key": self.primary_key,
        }


def random_id(length=10):
    import random, string

    return "".join(random.choices(string.ascii_letters + string.digits, k=length))


def dict_factory(cursor, row):
    d = {}
    for idx, col in enumerate(cursor.description):
        d[col[0]] = row[idx]
    return d


def transform_expression(
    exp: typing.Optional[str] = "",
    dict_params: typing.Dict[str, typing.Any] = {},
    type: str = "WHERE",
):
    param_key = r":\w+"
    list_params = []
    if exp:
        for match in re.findall(param_key, exp):
            list_params.append(dict_params[match[1:]])
            exp = exp.replace(match, "?")
        return f"{type} {exp}", list_params
    else:
        return "", list_params


class SqliteDB:
    db_path: str

    def __init__(self, db_path) -> None:
        self.db_path = str(db_path)

    def connect(self):
        is_uri = self.db_path.startswith("file://")
        conn = sqlite3.connect(self.db_path, uri=is_uri)
        return conn

    def create_table(self, table_name=f"table-{random_id()}", columns=[]):
        with self.connect() as conn:
            conn.execute(
                f"""
                CREATE TABLE "{table_name}" (
                    id INTEGER PRIMARY KEY NOT NULL,
                    created_at INTEGER DEFAULT CURRENT_TIMESTAMP,
                    updated_at INTEGER DEFAULT CURRENT_TIMESTAMP
                );
            """
            )

            for column in columns:
                self.create_column(table_name, column.get("name"), column.get("type"))

            return self.get_table(table_name)

    def update_table(self, name: str, changes: typing.Dict[str, str]):
        with self.connect() as conn:
            if "name" in changes:
                old_name = name
                new_name = changes["name"]

                conn.execute(
                    f"""
                    ALTER TABLE "{old_name}"
                    RENAME TO "{new_name}"
                """
                )

            return self.get_table(new_name)

    def get_table(self, name: str):
        with self.connect() as conn:
            [count] = conn.execute(
                f"SELECT count(*) FROM sqlite_master WHERE name=?", [name]
            ).fetchone()

            if count == 0:
                raise TableNotFound(name)
            else:
                return SqliteTable(self, name)

    def get_tables(self):
        with self.connect() as cur:
            rows = cur.execute(
                """
                SELECT name
                FROM sqlite_schema
                WHERE type='table'
                ORDER BY name
            """
            ).fetchall()

            return [SqliteTable(db=self, name=name) for (name,) in rows]

    def delete_table(self, name: str):
        cur = self.connect().cursor()
        cur.execute(
            f"""
            DROP TABLE "{name}"
        """
        )

    def duplicate_table(self, name: str):
        cur = self.connect().cursor()
        new_name = f"{name}-{random_id()}"
        cur.execute(
            f"""
            CREATE TABLE "{new_name}" AS
            SELECT * FROM "{name}"
        """
        )
        return SqliteTable(db=self, name=new_name)

    def db_types(self):
        return [
            {"type": "NULL", "input_spec": {"type": "hidden"}},
            {"type": "INTEGER", "input_spec": {"type": "number"}},
            {"type": "REAL", "input_spec": {"type": "number"}},
            {"type": "TEXT", "input_spec": {"type": "text"}},
            {"type": "BOOL", "input_spec": {"type": "checkbox"}},
            {"type": "BLOB", "input_spec": {"type": "file"}},
        ]

    def get_column(self, table_name: str, column_name: str) -> SqliteColumn:
        return self.get_table(table_name).get_column(column_name)

    def get_columns(self, table_name: str):
        cur = self.connect().cursor()
        rows = cur.execute(
            f"""
            pragma table_info("{table_name}")
        """
        ).fetchall()

        return [
            SqliteColumn(
                table=self.get_table(table_name),
                id=id,
                name=name,
                type=type,
                not_null=not_null == 1,
                default=default,
                primary_key=primary_key == 1,
            )
            for (id, name, type, not_null, default, primary_key) in rows
        ]

    def update_column(
        self, table_name: str, column_name: str, changes: typing.Dict[str, typing.Any]
    ):
        with self.connect() as conn:
            column = self.get_column(table_name, column_name)

            if ("type" in changes) or ("not_null" in changes) or ("default" in changes):
                new_column_name = changes.get("name", column_name)
                new_column_type = changes.get("type", column.type)
                new_column_not_null = changes.get("not_null", column.not_null)
                new_column_default = changes.get("default", column.default)

                if new_column_not_null:
                    not_null_exp = "NOT NULL"
                else:
                    not_null_exp = ""

                if new_column_default:
                    bias_exp = f"DEFAULT {new_column_default}"
                else:
                    bias_exp = ""

                uuid = random_id()
                temp_column_name = f"{new_column_name}_{uuid}"

                conn.executescript(
                    f"""
                    ALTER TABLE "{table_name}"
                    ADD COLUMN "{temp_column_name}" "{new_column_type}"
                    {bias_exp}
                    {not_null_exp};

                    UPDATE "{table_name}"
                    SET "{temp_column_name}" = "{column_name}";
                    
                    ALTER TABLE "{table_name}"
                    DROP COLUMN "{column_name}";
                    
                    ALTER TABLE "{table_name}"
                    RENAME COLUMN "{temp_column_name}" TO "{new_column_name}"
                """
                )
            elif "name" in changes:
                new_column_name = changes["name"]
                new_column_type = column.type
                conn.execute(
                    f"""
                    ALTER TABLE "{table_name}"
                    RENAME COLUMN "{column_name}" TO "{new_column_name}"
                """
                )
            else:
                raise Exception("Nothing changed")

            return self.get_column(table_name, new_column_name)

    def delete_column(self, table_name: str, column_name: str):
        cur = self.connect().cursor()

        cur.execute(
            f"""
            ALTER TABLE "{table_name}"
            DROP COLUMN "{column_name}"
        """
        )

    def create_column(
        self, table_name: str, column_name=f"column_{random_id()}", column_type=None
    ):
        cur = self.connect().cursor()

        if column_type is None:
            column_type = "TEXT"
            reference_exp = ""
        elif type(column_type) == str:
            reference_exp = ""
        elif type(column_type) == SqliteColumn:
            reference_exp = """
                REFERENCES "{table_name}" ("{column_name}")
            """
            column_type = column_type.type
        else:
            raise Exception(f"Invalid column type: {column_type}")

        cur.execute(
            f"""
            ALTER TABLE "{table_name}"
            ADD COLUMN "{column_name}" "{column_type}"
            {reference_exp}
        """
        )

        return self.get_column(table_name, column_name)


class SqliteTable:
    name: str
    bias: typing.Dict

    def __init__(
        self,
        db: SqliteDB,
        name: str,
        bias: typing.Dict = {},
    ) -> None:
        self.db = db
        self.name = name
        self.bias = bias

    @property
    def editor_dto(self):
        return {"name": self.name, "columns": [c.editor_dto for c in self.columns]}

    @property
    def columns(self):
        with self.db.connect() as conn:
            rows = conn.execute(
                f"""
                pragma table_info("{self.name}")
            """
            ).fetchall()

            if len(rows) == 0:
                raise TableNotFound(self.name)

            return [
                SqliteColumn(
                    table=self,
                    id=id,
                    name=name,
                    type=type,
                    not_null=not_null == 1,
                    default=default,
                    primary_key=primary_key == 1,
                )
                for (id, name, type, not_null, default, primary_key) in rows
            ]

    def add_bias(self, where_exp: str, where_params: typing.List[str]):
        if self.bias and len(self.bias) > 0:
            bias_exp = " AND ".join([f'"{col}" = ?' for col in self.bias.keys()])
            bias_params = list(self.bias.values())
            return where_exp + " AND " + bias_exp, where_params + bias_params
        else:
            return where_exp, where_params

    def get_column(self, column_name: str):
        return [c for c in self.columns if c.name == column_name][0]

    def select(
        self,
        where: typing.Optional[str] = None,
        columns: typing.Optional[typing.List[str]] = None,
        params: typing.Dict[str, typing.Any] = {},
    ) -> typing.List["SqliteRow"]:
        if not columns:
            columns = [c.name for c in self.columns]

        with self.db.connect() as conn:
            where_exp, list_params = transform_expression(where, params)
            where_exp, list_params = self.add_bias(where_exp, list_params)
            rows_exp = ", ".join(columns)
            query = f'SELECT {rows_exp} FROM "{self.name}" {where_exp}'
            rows = conn.execute(query, list_params).fetchall()
            return [SqliteRow(self, dict(zip(columns, row))) for row in rows]

    def select_one(
        self,
        where: typing.Optional[str] = None,
        columns: typing.Optional[typing.List[str]] = None,
        params: typing.Dict[str, typing.Any] = {},
    ) -> "SqliteRow":
        if not columns:
            columns = [c.name for c in self.columns]

        with self.db.connect() as conn:
            where_exp, list_params = transform_expression(where, params)
            rows_exp = ", ".join(columns)
            query = f'SELECT {rows_exp} FROM "{self.name}" {where_exp}'
            row = conn.execute(query, list_params).fetchone()
            return SqliteRow(self, dict(zip(columns, row)))

    def insert(
        self,
        values: typing.Union[
            typing.Dict[str, typing.Any], typing.List[typing.Dict[str, typing.Any]]
        ] = {},
        columns: typing.Optional[typing.List[str]] = None,
    ) -> typing.Union[typing.List["SqliteRow"], "SqliteRow"]:
        if columns is None:
            columns = [c.name for c in self.columns]

        if type(values) == dict:
            values_list = [values]
        elif type(values) == list:
            values_list = values
        else:
            raise Exception(f"Invalid values argument type: {values}")

        if len(values_list) == 0:
            return []
        elif len(values_list) > 1:
            keys = set(values_list[0].keys())
            for values_item in values_list:
                if set(values_item.keys()) != keys:
                    raise Exception("All dicts must have the same keys")

        columns_exp = ", ".join(columns)
        if len(values_list[0]) == 0:
            query = f'INSERT INTO "{self.name}" DEFAULT VALUES RETURNING {columns_exp}'
            params: typing.List[str] = []
        else:
            inserted_columns = values_list[0].keys()
            keys_exp = ", ".join(inserted_columns)
            value_exp = ", ".join("?" for _ in inserted_columns)
            values_exp = ", ".join(f"({value_exp})" for _ in values_list)
            params = []
            for value_item in values_list:
                params.extend(value_item[col] for col in inserted_columns)
            query = f'INSERT INTO "{self.name}" ({keys_exp}) VALUES {values_exp} RETURNING {columns_exp}'

        with self.db.connect() as conn:
            result = conn.execute(query, params).fetchall()

            if type(values) == dict:
                return SqliteRow(self, dict(zip(columns, result[0])))
            elif type(values) == list:
                return [SqliteRow(self, dict(zip(columns, row))) for row in result]
            else:
                raise Exception(f"Invalid values argument type: {values}")

    def update(
        self,
        where: str,
        set: typing.Dict[str, typing.Any] = {},
        columns: typing.Optional[typing.List[str]] = None,
        params: typing.Dict[str, typing.Any] = {},
    ) -> typing.List["SqliteRow"]:
        if not columns:
            columns = [c.name for c in self.columns]

        with self.db.connect() as conn:
            rows_exp = ", ".join(columns)

            set_exp = ", ".join([f"{key} = ?" for key in set.keys()])
            set_params = [value for value in set.values()]
            where_exp, where_params = transform_expression(where, params)
            where_exp, where_params = self.add_bias(where_exp, where_params)
            query = (
                f'UPDATE "{self.name}" SET {set_exp} {where_exp} RETURNING {rows_exp}'
            )
            result = conn.execute(query, set_params + where_params).fetchall()

            return [SqliteRow(self, dict(zip(columns, row))) for row in result]

    def delete(
        self,
        where: str,
        columns: typing.Optional[typing.List[str]] = None,
        params: typing.Dict[str, typing.Any] = {},
    ) -> typing.List["SqliteRow"]:
        if not columns:
            columns = [c.name for c in self.columns]
        with self.db.connect() as conn:
            rows_exp = ", ".join(columns)

            where_exp, list_params = transform_expression(where, params)
            where_exp, list_params = self.add_bias(where_exp, list_params)
            query = f'DELETE FROM "{self.name}" {where_exp} RETURNING {rows_exp}'
            result = conn.execute(query, list_params).fetchall()

            return [SqliteRow(self, dict(zip(columns, row))) for row in result]

    def join(self, table, on: str, params: typing.Dict = {}) -> "SqliteJoin":
        if type(table) == str:
            right_table = self.db.get_table(table)
        elif type(table) == SqliteTable:
            right_table = table
        else:
            raise InvalidTableJoin(table)
        return SqliteJoin(
            db=self.db,
            left_table=self,
            right_table=right_table,
            on_exp=on,
            on_params=params,
        )


class SqliteJoin:
    def __init__(
        self,
        db: SqliteDB,
        left_table: SqliteTable,
        right_table: SqliteTable,
        on_exp: str,
        on_params: typing.Dict,
    ):
        self.db = db
        self.left_table = left_table
        self.right_table = right_table
        self.on_exp = on_exp
        self.on_params = on_params

    def select(
        self,
        where: typing.Optional[str] = None,
        columns: typing.Optional[typing.List[str]] = None,
        params: typing.Dict[str, typing.Any] = {},
    ):
        if not columns:
            columns = [
                f'"{col.name}" as "{self.left_table.name}"."{col.name}"'
                for col in self.left_table.columns
            ]
            columns += [
                f'"{col.name}" as "{self.right_table.name}"."{col.name}"'
                for col in self.right_table.columns
            ]
        with self.db.connect() as conn:
            where_exp, where_params = transform_expression(where, params)
            on_exp, on_params = transform_expression(
                exp=self.on_exp, dict_params=self.on_params, type="ON"
            )
            list_params = [*on_params, *where_params]
            rows_exp = ", ".join(columns)
            query = f"""
            SELECT {rows_exp}
            FROM "{self.left_table.name}"
            JOIN "{self.right_table.name}"
            {on_exp}
            {where_exp}
            """

            rows = conn.execute(query, list_params).fetchall()

            return [SqliteRow(self, dict(zip(columns, row))) for row in rows]


class SqliteRow:
    def __init__(self, table: typing.Union[SqliteTable, SqliteJoin], data: typing.Dict):
        self.table = table
        self.initial_state = data
        self.changes: typing.Dict = {}

    def get(self, column_idx):
        if type(column_idx) == str:
            column_name = column_idx
        elif type(column_idx) == int:
            column_name = [c.name for c in self.table.columns][column_idx]
        if column_name in self.changes:
            return self.changes[column_name]
        elif column_name in self.initial_state:
            return self.initial_state[column_name]
        elif column_name in [c.name for c in self.table.columns]:
            return None
        else:
            raise Exception(f"Column '{column_name}' not found")

    def set(self, column_name: str, value: typing.Any):
        self.changes[column_name] = value

    def __getitem__(self, key):
        return self.get(key)

    def __setitem__(self, key, value):
        self.set(key, value)

    @property
    def identifier_selector(self):
        return " AND ".join(
            [f"{c.name} = :{c.name}" for c in self.table.columns if c.primary_key]
        )

    @property
    def identifier_params(self):
        return {c.name: self.get(c.name) for c in self.table.columns if c.primary_key}

    def assert_all_identifiers_found(self):
        valid = all(
            [c.name in self.initial_state for c in self.table.columns if c.primary_key]
        )
        if not valid:
            raise MissingColumns(
                [
                    c.name
                    for c in self.table.columns
                    if c.primary_key and c.name not in self.initial_state
                ]
            )

    def save(self):
        self.assert_all_identifiers_found()
        self.table.update(
            where=self.identifier_selector,
            set=self.changes,
            params=self.identifier_params,
        )
        self.initial_state.update(self.changes)
        self.changes = {}

    def delete(self):
        self.assert_all_identifiers_found()
        self.table.delete(
            where=self.identifier_selector,
            params=self.identifier_params,
        )

    @property
    def __dict__(self):
        return {**self.initial_state, **self.changes}

    def __iter__(self):
        return iter(self.__dict__.items())

    def __repr__(self) -> str:
        return f"<SqliteRow {self.__dict__}>"



import os, tempfile, urllib.request, zipfile, requests
from uuid import uuid4 as uuid
from pathlib import Path
from .credentials import get_credentials
from .utils.file import files_from_directory

CLOUD_API_ENDPOINT = (
    os.environ.get("CLOUD_API_ENDPOINT") or "https://cloud-api.abstra.cloud"
)


def _generate_zip_file(root_path: Path) -> Path:
    zip_path = Path(tempfile.gettempdir(), f"{uuid()}.zip")
    with zipfile.ZipFile(zip_path, "w") as zip_file:
        for file in files_from_directory(root_path):
            zip_file.write(file, file.relative_to(root_path))
    return zip_path


def _create_build(headers: dict) -> dict:
    return requests.post(f"{CLOUD_API_ENDPOINT}/cli/builds", headers=headers).json()


def _upload_file(url: str, file_path: Path, headers: dict):
    with file_path.open("rb") as f:
        req = urllib.request.Request(url=url, method="PUT", data=f.read())
        urllib.request.urlopen(req)


def _update_build(build_id: str, headers: dict):
    requests.patch(
        f"{CLOUD_API_ENDPOINT}/cli/builds/{build_id}",
        headers=headers,
    )


def deploy(workspace_root: Path = Path(".")):
    cwd = Path.cwd()
    root_path = (cwd / workspace_root).resolve()
    credentials = get_credentials(root_path)

    if not credentials:
        print("No API token found. Please login with `abstra login`")
        return

    headers = {"Api-Authorization": f"Bearer {credentials}"}
    data = _create_build(headers)
    url = data["url"]
    build_id = data["buildId"]
    zip_path = _generate_zip_file(root_path)
    _upload_file(url=url, file_path=zip_path, headers=headers)

    _update_build(headers=headers, build_id=build_id)



from .socket import send, receive


def execute_action(type, payload):
    send(
        {
            "type": "action",
            "action": {
                "type": type,
                **payload,
            },
        }
    )

    response = receive()

    if "errorMessage" in response:
        raise Exception(response["errorMessage"])

    return response.get("value", None)


def redirect(url):
    return execute_action("redirect", {"url": url})


def execute_js(code, context={}):
    return execute_action("execute-js", {"code": code, "context": context})



import uuid
from ..apis import main as api_main


class MembershipModel:
    id: uuid.UUID
    name: str
    email: str

    def __init__(self, id: str, name: str, email: str, picture: str):
        self.id = uuid.UUID(id)
        self.name = name
        self.email = email
        self.picture = picture


def list_memberships():
    api_data = api_main.hf_api_runner("GET", "members")
    return [
        MembershipModel(
            id=d["id"], name=d["name"], email=d["email"], picture=d["picture"]
        )
        for d in api_data
    ]


def add_membership(email: str):
    api_main.hf_api_runner("POST", "member", {"email": email})



import datetime, re
from ..widget_base import Input
from typing import Optional, Union


class TimeInput(Input):
    type = "time-input"

    def __init__(self, key: str, label: str, **kwargs):
        super().__init__(key)
        self.label = label
        self.initial_value = kwargs.get("initial_value", "")
        self.required = kwargs.get("required", True)
        self.hint = kwargs.get("hint", None)
        self.columns = kwargs.get("columns", 1)
        self.full_width = kwargs.get("full_width", False)
        self.format = kwargs.get("format", "24hs")
        self.disabled = kwargs.get("disabled", False)

    def json(self, **kwargs):
        return {
            "type": self.type,
            "key": self.key,
            "label": self.label,
            "format": self.format,
            "hint": self.hint,
            "initialValue": TimeInput.__revert_value(self.initial_value),
            "required": self.required,
            "columns": self.columns,
            "fullWidth": self.full_width,
            "disabled": self.disabled,
        }

    @staticmethod
    def __revert_value(
        value: Optional[Union[datetime.time, str, dict]]
    ) -> Optional[dict]:
        if isinstance(value, str) and re.match("^\\d{2}:\\d{2}$", value):
            return {
                "hour": int(value.split(":")[0]),
                "minute": int(value.split(":")[1]),
            }
        if isinstance(value, datetime.time):
            return {"hour": value.hour, "minute": value.minute}
        return None

    @staticmethod
    def __convert_answer(answer) -> Optional[datetime.time]:
        return datetime.time(answer["hour"], answer["minute"]) if answer else None

    def convert_answer(self, answer) -> Optional[datetime.time]:
        """
        Returns:
            datetime.time: A datetime.time object representing the value entered by the user
        """
        return self.__convert_answer(answer)
